#include <tcl.h>
#include <bindata.h>
#include <math.h>

int
Tkgraph_LlimitCmd(ClientData clientData, Tcl_Interp *interp, int argc, char **argv)
{
    int		xyflag = 0,
    		minflag = 1,
		maxflag = 1;
    int		i;
    Bin_Object *objPtr = (Bin_Object *)NULL;
    double	x, y,
    		xmin, ymin,
		xmax, ymax;
    char	buf[TCL_DOUBLE_SPACE];
    int		targc = argc;
    int		largc = argc;
    int		nelem = argc;
    char	**targv = argv;
    char	**largv = argv;

    ++targv, --targc;

    largv = targv;
    largc = targc;
    while (targc && **targv == '-') {
	if (!strcmp(*targv, "-xy")) {
	    xyflag = 1;
	} else if (!strcmp(*targv, "-min")) {
	    minflag = 1, maxflag = 0;
	} else if (!strcmp(*targv, "-max")) {
	    maxflag = 1, minflag = 0;
	} else {
	    /* We found data.  set targc to 1 so we exit out of the loop */
	    /*
	    Tcl_AppendResult(interp, "Usage: \"", *argv,
		"?-xy? ?-min? ?-max? list\"", (char *)NULL);
	    return TCL_ERROR;
	    */
	    break;
	}
	++targv, --targc;
    }

    if (targc == 1) {
	if (Tcl_SplitList(interp, argv[0], &argc, &largv) != TCL_OK) {
	    if (largv != NULL) {
		ckfree((char *) largv);
	    }
	    return TCL_ERROR;
	}

	/* If we have one argument, then assume that it's a Bin_Object object.
	 * Create the coordinate array from this object.  The idea is that
	 * we want to be able to pass the graph a pre-existing binary
	 * structure containing the coordinates so that creating a new
	 * dataset does not involve billions of calls to Tcl_GetDouble().
	 */

	if (targc == 1) {
	    objPtr = Bin_GetObject(interp, *targv);
	    if (objPtr == (Bin_Object *)NULL) {
		return TCL_ERROR;
	    }

	    if (strcmp(objPtr->type, "float*") != 0
		&& strcmp(objPtr->type, "double*") != 0
		&& strcmp(objPtr->type, "int*") != 0
		&& strcmp(objPtr->type, "long*") != 0) {
		Tcl_AppendResult(interp, "Bad object type.  Must be one of double*, float*, int*, or long*", (char *)NULL);
		return TCL_ERROR;
	    }

	    nelem = objPtr->size / objPtr->elemsize;
	}
    } else {
	nelem = targc;
    }
    if (targc & 1 && (objPtr == (Bin_Object *)NULL)) {
	Tcl_AppendResult(interp,
		"odd number of coordinates specified for line",
		(char *) NULL);
	return TCL_ERROR;
    }

    if ((nelem < 1) || (xyflag && nelem < 2) || (xyflag && nelem%2)) {
	Tcl_AppendResult(interp, "wrong # args: should be \"", *argv,
		"?-xy? ?-min? ?-max? list\"", (char *)NULL);
	return TCL_ERROR;
    }

    if (objPtr == NULL) {
	if (Tcl_GetDouble(interp, targv[0], &xmin) != TCL_OK)
	    return TCL_ERROR;
	if (xyflag) {
	    if (Tcl_GetDouble(interp, targv[1], &ymin) != TCL_OK)
		return TCL_ERROR;
	}
    } else {
	switch(*objPtr->type) {
	    case 'i': xmin = ((int *)objPtr->data)[0];
	              ymin = ((int *)objPtr->data)[1];
		      break;
	    case 'l': xmin = ((long *)objPtr->data)[0];
	              ymin = ((long *)objPtr->data)[1];
		      break;
	    case 'f': xmin = ((float *)objPtr->data)[0];
	              ymin = ((float *)objPtr->data)[1];
		      break;
	    case 'd': xmin = ((double *)objPtr->data)[0];
	              ymin = ((double *)objPtr->data)[1];
		      break;
	}
    }
    xmax = xmin;
    ymax = ymin;

    for (i = 0; i < nelem; i++) {
	if (!objPtr) {
	    if (Tcl_GetDouble(interp, targv[i], &x) != TCL_OK)
		return TCL_ERROR;
	    xmin = (x < xmin)?x:xmin;
	    xmax = (x > xmax)?x:xmax;

	    if (xyflag) {
		if (Tcl_GetDouble(interp, targv[i+1], &y) != TCL_OK)
		    return TCL_ERROR;
		ymin = (y < ymin)?y:ymin;
		ymax = (y > ymax)?y:ymax;
		i++;
	    }
	} else {
	    /*
	    printf("Checking %g, %g against %d\n", xmin, xmax, ((int *)objPtr->data)[i]);
	    */
	    switch (*objPtr->type) {
		case 'i':   if (((int *)objPtr->data)[i] < xmin)
				xmin = (((int *)objPtr->data)[i]);
			    if (((int *)objPtr->data)[i] > xmax)
				xmax = (((int *)objPtr->data)[i]);
			    break;
		case 'l':   if (((long *)objPtr->data)[i] < xmin)
				xmin = (((long *)objPtr->data)[i]);
			    if (((long *)objPtr->data)[i] > xmax)
				xmax = (((long *)objPtr->data)[i]);
			    break;
		case 'f':   if (((float *)objPtr->data)[i] < xmin)
				xmin = (((float *)objPtr->data)[i]);
			    if (((float *)objPtr->data)[i] > xmax)
				xmax = (((float *)objPtr->data)[i]);
			    break;
		case 'd':   if (((double *)objPtr->data)[i] < xmin)
				xmin = (((double *)objPtr->data)[i]);
			    if (((double *)objPtr->data)[i] > xmax)
				xmax = (((double *)objPtr->data)[i]);
			    break;
	    }
	    if (xyflag) {
		switch (*objPtr->type) {
		    case 'i':   if (((int *)objPtr->data)[i+1] < ymin)
				    ymin = (((int *)objPtr->data)[i+1]);
				if (((int *)objPtr->data)[i+1] > ymax)
				    ymax = (((int *)objPtr->data)[i+1]);
				break;
		    case 'l':   if (((long *)objPtr->data)[i+1] < ymin)
				    ymin = (((long *)objPtr->data)[i+1]);
				if (((long *)objPtr->data)[i+1] > ymax)
				    ymax = (((long *)objPtr->data)[i+1]);
				break;
		    case 'f':   if (((float *)objPtr->data)[i+1] < ymin)
				    ymin = (((float *)objPtr->data)[i+1]);
				if (((float *)objPtr->data)[i+1] > ymax)
				    ymax = (((float *)objPtr->data)[i+1]);
				break;
		    case 'd':   if (((double *)objPtr->data)[i+1] < ymin)
				    ymin = (((double *)objPtr->data)[i+1]);
				if (((double *)objPtr->data)[i+1] > ymax)
				    ymax = (((double *)objPtr->data)[i+1]);
				break;
		}
		i++;
	    }
	}
    }

    if (minflag) {
	Tcl_PrintDouble(interp, xmin, buf);
	Tcl_AppendElement(interp, buf);
	if (xyflag) {
	    Tcl_PrintDouble(interp, ymin, buf);
	    Tcl_AppendElement(interp, buf);
	}
    }
    if (maxflag) {
	Tcl_PrintDouble(interp, xmax, buf);
	Tcl_AppendElement(interp, buf);
	if (xyflag) {
	    Tcl_PrintDouble(interp, ymax, buf);
	    Tcl_AppendElement(interp, buf);
	}
    }

    return TCL_OK;
}

int
Tkgraph_TickroundCmd(ClientData clientData, Tcl_Interp *interp, int argc, char **argv) {
    double a = 0.0;
    double b = 0.0;
    double t = 0.0;
    double l1 = 0.0;
    double l2 = 0.0;
    double nticks = 0.0;
    double interval = 0.0;
    double max = 0.0;
    double min = 0.0;
    double mantissa = 0.0;

    char result[TCL_DOUBLE_SPACE];

    if (argc != 3) {
	Tcl_AppendResult(interp, "Bad number of arguments.  Should be:  ", *argv, " a b", (char *)NULL);
	return TCL_ERROR;
    }

    if (Tcl_GetDouble(interp, argv[1], &a) != TCL_OK)
	return TCL_ERROR;
    if (Tcl_GetDouble(interp, argv[2], &b) != TCL_OK)
	return TCL_ERROR;

    if (a > b) {
	t = a;
	a = b;
	b = t;
    }

    if (a == b) {
	Tcl_PrintDouble(interp, a, result);
	Tcl_AppendElement(interp, result);
	Tcl_PrintDouble(interp, b, result);
	Tcl_AppendElement(interp, result);
	Tcl_AppendElement(interp, "0");
	Tcl_AppendElement(interp, "0");
	return TCL_OK;
    }

    l1 = log((b - a) / 2.0) / log(10.);
    l2 = 1.0;
    mantissa = l1 - floor(l1);

    if (mantissa > .87)
	l2 = 4.;
    else if (mantissa > .69897)
	l2 = 2.;
    else if (mantissa < .39794)
	l2 = .5;
    else if (mantissa < .2)
	l2 = .25;

    interval = pow(10, (double)(floor(l1))) * l2;
    max = ceil( (b + interval)/interval) * interval;
    min = ceil( (a - interval)/interval) * interval;
    nticks = ceil((max - min) / interval);

    if (min == -0.0) min = 0.0;

    Tcl_PrintDouble(interp, min, result);
    Tcl_AppendElement(interp, result);
    Tcl_PrintDouble(interp, max, result);
    Tcl_AppendElement(interp, result);
    Tcl_PrintDouble(interp, interval, result);
    Tcl_AppendElement(interp, result);
    Tcl_PrintDouble(interp, nticks, result);
    Tcl_AppendElement(interp, result);
    return TCL_OK;
}
